# coding=utf-8
"""
HuaweiDNS API
华为DNS解析操作库
https://support.huaweicloud.com/api-dns/zh-cn_topic_0037134406.html
@author: cybmp3
"""


from hashlib import sha256
from hmac import new as hmac
from binascii import hexlify
from json import loads as jsondecode, dumps as jsonencode
from logging import debug, info, warning
from datetime import datetime

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode

__author__ = 'New Future'
BasicDateFormat = "%Y%m%dT%H%M%SZ"
Algorithm = "SDK-HMAC-SHA256"


# __all__ = ["request", "ID", "TOKEN", "PROXY"]


class Config:
    ID = "id"  # AK
    TOKEN = "TOKEN"  # AS
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SCHEME = 'https'
    SITE = 'dns.myhuaweicloud.com'  # API endpoint


def HexEncodeSHA256Hash(data):
    sha = sha256()
    sha.update(data)
    return sha.hexdigest()


def StringToSign(canonical_request, t):
    b = HexEncodeSHA256Hash(canonical_request)
    return "%s\n%s\n%s" % (Algorithm, datetime.strftime(t, BasicDateFormat), b)


def CanonicalHeaders(headers, signed_headers):
    a = []
    __headers = {}
    for key in headers:
        key_encoded = key.lower()
        value = headers[key]
        value_encoded = value.strip()
        __headers[key_encoded] = value_encoded
    for key in signed_headers:
        a.append(key + ":" + __headers[key])
    return '\n'.join(a) + "\n"


def request(method, path, param=None, body=None, **params):
    # path 是不带host但是 前面需要带 / , body json 字符串或者自己从dict转换下
    # 也可以自己改成 判断下是不是post 是post params就是body
    if param:
        params.update(param)

    query = urlencode(sorted(params.items()))
    headers = {"content-type": "application/json"}  # 初始化header
    headers["X-Sdk-Date"] = datetime.strftime(
        datetime.utcnow(), BasicDateFormat)
    headers["host"] = API.SITE
    # 如何后来有需要把header头 key转换为小写 value 删除前导空格和尾随空格
    sign_headers = []
    for key in headers:
        sign_headers.append(key.lower())
    # 先排序
    sign_headers.sort()

    if body is None:
        body = ""

    hex_encode = HexEncodeSHA256Hash(body.encode('utf-8'))
    # 生成文档中的CanonicalRequest
    canonical_headers = CanonicalHeaders(headers, sign_headers)

    # 签名中的path 必须 / 结尾
    if path[-1] != '/':
        sign_path = path + "/"
    else:
        sign_path = path

    canonical_request = "%s\n%s\n%s\n%s\n%s\n%s" % (method.upper(), sign_path, query,
                                                    canonical_headers, ";".join(sign_headers), hex_encode)

    hashed_canonical_request = HexEncodeSHA256Hash(
        canonical_request.encode('utf-8'))

    # StringToSign
    str_to_sign = "%s\n%s\n%s" % (
        Algorithm, headers['X-Sdk-Date'], hashed_canonical_request)

    secret = Config.TOKEN
    # 计算签名  HexEncode(HMAC(Access Secret Key, string to sign))
    signature = hmac(secret.encode(
        'utf-8'), str_to_sign.encode('utf-8'), digestmod=sha256).digest()
    signature = hexlify(signature).decode()
    # 添加签名信息到请求头
    auth_header = "%s Access=%s, SignedHeaders=%s, Signature=%s" % (
        Algorithm, Config.ID, ";".join(sign_headers), signature)
    headers['Authorization'] = auth_header
    # 创建Http请求

    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)
    conn.request(method, API.SCHEME + "://" + API.SITE +
                 path + '?' + query, body, headers)
    info(API.SCHEME + "://" + API.SITE + path + '?' + query, body)
    resp = conn.getresponse()
    data = resp.read().decode('utf8')
    resp.close()
    if resp.status < 200 or resp.status >= 300:

        warning('%s : error[%d]: %s', path, resp.status, data)
        raise Exception(data)
    else:
        data = jsondecode(data)
        debug('%s : result:%s', path, data)
        return data


def get_zone_id(domain):
    """
    切割域名获取主域名和对应ID https://support.huaweicloud.com/api-dns/dns_api_62003.html
    优先匹配级数最长的主域名
    """
    zoneid = None
    domain_slice = domain.split('.')
    index = len(domain_slice)
    root_domain = '.'.join(domain_slice[-2:])
    zones = request('GET', '/v2/zones', limit=500, name=root_domain)['zones']
    while (not zoneid) and (index >= 2):
        domain = '.'.join(domain_slice[-index:]) + '.'
        zone = next((z for z in zones if domain == (z.get('name'))), None)
        zoneid = zone and zone['id']
        index -= 1
    return zoneid


def get_records(zoneid, **conditions):
    """
        获取记录ID
        返回满足条件的所有记录[]
        https://support.huaweicloud.com/api-dns/dns_api_64004.html
        TODO 大于500翻页
    """
    cache_key = zoneid + "_" + \
        conditions.get('name', "") + "_" + conditions.get('type', "")
    if not hasattr(get_records, 'records'):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ('id', 'type', 'name', 'records', 'ttl')

    if zoneid not in get_records.records:
        get_records.records[cache_key] = {}

        data = request('GET', '/v2/zones/' + zoneid + '/recordsets',
                       limit=500, **conditions)

        # https://{DNS_Endpoint}/v2/zones/2c9eb155587194ec01587224c9f90149/recordsets?limit=&offset=
        if data:
            for record in data['recordsets']:
                info(record)
                get_records.records[cache_key][record['id']] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}
    records = {}
    for (zid, record) in get_records.records[cache_key].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[zid] = record
    return records


def update_record(domain, value, record_type='A'):
    """
        更新记录
        update
        https://support.huaweicloud.com/api-dns/UpdateRecordSet.html
        add
        https://support.huaweicloud.com/api-dns/dns_api_64001.html
    """
    info(">>>>>%s(%s)", domain, record_type)
    zoneid = get_zone_id(domain)
    if not zoneid:
        raise Exception("invalid domain: [ %s ] " % domain)
    domain += '.'
    records = get_records(zoneid, name=domain, type=record_type)
    cache_key = zoneid + "_" + domain + "_" + record_type
    result = {}
    if records:  # update
        for (rid, record) in records.items():
            if record['records'] != value:
                """
                PUT https://{endpoint}/v2/zones/{zone_id}/recordsets/{recordset_id}

                {
                    "name" : "www.example.com.",
                    "description" : "This is an example record set.",
                    "type" : "A",
                    "ttl" : 3600,
                    "records" : [ "192.168.10.1", "192.168.10.2" ]
                }
                """
                body = {
                    "name": domain,
                    "description": "Managed by DDNS.",
                    "type": record_type,
                    "records": [
                        value
                    ]
                }
                # 如果TTL不为空，则添加到字典中
                if Config.TTL is not None:
                    body['ttl'] = Config.TTL
                res = request('PUT', '/v2/zones/' + zoneid + '/recordsets/' + record['id'],
                              body=str(jsonencode(body)))
                if res:
                    get_records.records[cache_key][rid]['records'] = value
                    result[rid] = res.get("name")
                else:
                    result[rid] = "Update fail!\n" + str(res)
            else:
                result[rid] = domain
    else:  # create
        body = {
            "name": domain,
            "description": "Managed by DDNS.",
            "type": record_type,
            "records": [
                value
            ]
        }
        # 如果TTL不为空，则添加到字典中
        if Config.TTL is not None:
            body['ttl'] = Config.TTL
        res = request('POST', '/v2/zones/' + zoneid + '/recordsets',
                      body=str(jsonencode(body)))
        if res:
            get_records.records[cache_key][res['id']] = res
            result = res
        else:
            result = domain + " created fail!"
    return result



# coding=utf-8
"""
DNSPOD API
DNSPOD 接口解析操作库
http://www.dnspod.com/docs/domains.html
@author: New Future
"""

from dns.dnspod import *  # noqa: F403

API.SITE = "api.dnspod.com"  # noqa: F405
API.DEFAULT = "default"  # noqa: F405

# https://github.com/NewFuture/DDNS/issues/286
# API.TOKEN_PARAM = "user_token"  # noqa: F405



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
DDNS
@author: New Future
@modified: rufengsuixing
"""
from __future__ import print_function
from time import ctime, asctime
from os import path, environ, name as os_name
from tempfile import gettempdir
from logging import DEBUG, basicConfig, info, warning, error, debug
from subprocess import check_output

import sys

from util import ip
from util.cache import Cache
from util.config import init_config, get_config

__version__ = "v3.1.1@2025-04-09T12:33:52+00:00"  # CI 时会被Tag替换
__description__ = "automatically update DNS records to dynamic local IP [自动更新DNS记录指向本地IP]"
__doc__ = """
ddns[%s]
(i) homepage or docs [文档主页]: https://ddns.newfuture.cc/
(?) issues or bugs [问题和帮助]: https://github.com/NewFuture/DDNS/issues
Copyright (c) New Future (MIT License)
""" % (__version__)

environ["DDNS_VERSION"] = "v3.1.1"

if getattr(sys, 'frozen', False):
    # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-OpenSSL-Certificate
    environ['SSL_CERT_FILE'] = path.join(
        getattr(sys, '_MEIPASS'), 'lib', 'cert.pem')


def get_ip(ip_type, index="default"):
    """
    get IP address
    """
    # CN: 捕获异常
    # EN: Catch exceptions
    value = None
    try:
        debug(f"get_ip({ip_type}, {index})")
        if index is False:  # disabled
            return False
        elif type(index) is list:  # 如果获取到的规则是列表，则依次判断列表中每一个规则，直到找到一个可以正确获取到的IP
            for i in index:
                value = get_ip(ip_type, i)
                if value:
                    break
        elif str(index).isdigit():  # 数字 local eth
            value = getattr(ip, "local_v" + ip_type)(index)
        elif index.startswith('cmd:'):  # cmd
            value = str(check_output(index[4:]).strip().decode('utf-8'))
        elif index.startswith('shell:'):  # shell
            value = str(check_output(
                index[6:], shell=True).strip().decode('utf-8'))
        elif index.startswith('url:'):  # 自定义 url
            value = getattr(ip, "public_v" + ip_type)(index[4:])
        elif index.startswith('regex:'):  # 正则 regex
            value = getattr(ip, "regex_v" + ip_type)(index[6:])
        else:
            value = getattr(ip, index + "_v" + ip_type)()
    except Exception as e:
        error(e)
    finally:
        return value


def change_dns_record(dns, proxy_list, **kw):
    for proxy in proxy_list:
        if not proxy or (proxy.upper() in ['DIRECT', 'NONE']):
            dns.Config.PROXY = None
        else:
            dns.Config.PROXY = proxy
        record_type, domain = kw['record_type'], kw['domain']
        print('\n%s %s(%s) ==> %s [via %s]' %
              (asctime(), domain, record_type, kw['ip'], proxy))
        try:
            return dns.update_record(domain, kw['ip'], record_type=record_type)
        except Exception as e:
            error(e)
    return False


def update_ip(ip_type, cache, dns, proxy_list):
    """
    更新IP
    """
    ipname = 'ipv' + ip_type
    domains = get_config(ipname)
    if not domains:
        return None
    if not isinstance(domains, list):
        domains = domains.strip('; ').replace(
            ',', ';').replace(' ', ';').split(';')
    index_rule = get_config('index' + ip_type, "default")  # 从配置中获取index配置
    address = get_ip(ip_type, index_rule)
    if not address:
        error('Fail to get %s address!', ipname)
        return False
    elif cache and (address == cache[ipname]):
        print('.', end=" ")  # 缓存命中
        return True
    record_type = (ip_type == '4') and 'A' or 'AAAA'
    update_fail = False  # https://github.com/NewFuture/DDNS/issues/16
    for domain in domains:
        domain = domain.lower()  # https://github.com/NewFuture/DDNS/issues/431
        if change_dns_record(dns, proxy_list, domain=domain, ip=address, record_type=record_type):
            update_fail = True
    if cache is not False:
        # 如果更新失败删除缓存
        cache[ipname] = update_fail and address


def main():
    """
    更新
    """
    init_config(__description__, __doc__, __version__)
    # Dynamicly import the dns module as configuration
    dns_provider = str(get_config('dns', 'dnspod').lower())
    dns = getattr(__import__('dns', fromlist=[dns_provider]), dns_provider)
    dns.Config.ID = get_config('id')
    dns.Config.TOKEN = get_config('token')
    dns.Config.TTL = get_config('ttl')
    if get_config('debug'):
        ip.DEBUG = get_config('debug')
        basicConfig(
            level=DEBUG,
            format='%(asctime)s <%(module)s.%(funcName)s> %(lineno)d@%(pathname)s \n[%(levelname)s] %(message)s')
        print("DDNS[", __version__, "] run:", os_name, sys.platform)
        if get_config("config"):
            print("Configuration was loaded from <==",
                  path.abspath(get_config("config")))
        print("=" * 25, ctime(), "=" * 25, sep=' ')

    proxy = get_config('proxy') or 'DIRECT'
    proxy_list = proxy if isinstance(
        proxy, list) else proxy.strip('; ').replace(',', ';').split(';')

    cache_config = get_config('cache', True)
    if cache_config is False:
        cache = cache_config
    elif cache_config is True:
        cache = Cache(path.join(gettempdir(), 'ddns.cache'))
    else:
        cache = Cache(cache_config)

    if cache is False:
        info("Cache is disabled!")
    elif get_config("config_modified_time") is None or get_config("config_modified_time") >= cache.time:
        warning("Cache file is out of dated.")
        cache.clear()
    else:
        debug("Cache is empty.")
    update_ip('4', cache, dns, proxy_list)
    update_ip('6', cache, dns, proxy_list)


if __name__ == '__main__':
    if sys.version_info.major == 3 and os_name == 'nt':
        from io import TextIOWrapper
        sys.stdout = TextIOWrapper(sys.stdout.detach(), encoding='utf-8')
        sys.stderr = TextIOWrapper(sys.stderr.detach(), encoding='utf-8')
    main()



# coding=utf-8
"""
Custom Callback API
自定义回调接口解析操作库

@author: 老周部落
"""

from json import loads as jsondecode
from logging import debug, info, warning
from time import time

try:
    # python 2
    from httplib import HTTPSConnection, HTTPConnection
    from urlparse import urlparse, parse_qsl
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection, HTTPConnection
    from urllib.parse import urlencode, urlparse, parse_qsl

__author__ = '老周部落'


class Config:
    ID = None  # 自定义回调 URL
    TOKEN = None  # 使用 JSON 编码的 POST 参数
    PROXY = None   # 代理设置
    TTL = None


def request(method, action, param=None, **params):
    """
    发送请求数据
    """
    if param:
        params.update(param)

    URLObj = urlparse(Config.ID)
    params = dict((k, params[k]) for k in params if params[k] is not None)
    info("%s/%s : %s", URLObj.netloc, action, params)

    if Config.PROXY:
        if URLObj.netloc == "http":
            conn = HTTPConnection(Config.PROXY)
        else:
            conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(URLObj.netloc, URLObj.port)
    else:
        if URLObj.netloc == "http":
            conn = HTTPConnection(URLObj.netloc, URLObj.port)
        else:
            conn = HTTPSConnection(URLObj.netloc, URLObj.port)

    headers = {}

    if method == "GET":
        if params:
            action += '?' + urlencode(params)
        params = ""
    else:
        headers["Content-Type"] = "application/x-www-form-urlencoded"

    params = urlencode(params)

    conn.request(method, action, params, headers)
    response = conn.getresponse()
    res = response.read().decode('utf8')
    conn.close()
    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]:%s', action, response.status, res)
        raise Exception(res)
    else:
        debug('%s : result:%s', action, res)
        return res


def replace_params(domain, record_type, ip, params):
    """
    替换定义常量为实际值
    """
    dict = {"__DOMAIN__": domain, "__RECORDTYPE__": record_type,
            "__TTL__": Config.TTL, "__TIMESTAMP__": time(), "__IP__": ip}
    for key, value in params.items():
        if dict.get(value):
            params[key] = dict.get(value)
    return params


def update_record(domain, value, record_type="A"):
    """
    更新记录
    """
    info(">>>>>%s(%s)", domain, record_type)

    result = {}

    if not Config.TOKEN:  # 此处使用 TOKEN 参数透传 POST 参数所用的 JSON
        method = "GET"
        URLObj = urlparse(Config.ID)
        path = URLObj.path
        query = dict(parse_qsl(URLObj.query))
        params = replace_params(domain, record_type, value, query)
    else:
        method = "POST"
        URLObj = urlparse(Config.ID)
        path = URLObj.path
        params = replace_params(domain, record_type,
                                value, jsondecode(Config.TOKEN))

    res = request(method, path, params)

    if res:
        result = "Callback Request Success!\n" + res
    else:
        result = "Callback Request Fail!\n"

    return result



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from argparse import ArgumentParser, ArgumentTypeError, Namespace, RawTextHelpFormatter  # noqa: F401
from json import load as loadjson, dump as dumpjson
from logging import error
from os import stat, environ
from time import time

import sys


__cli_args = {}  # type: Namespace
__config = {}  # type: dict


def str2bool(v):
    """
    parse string to boolean
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')


def init_config(description, doc, version):
    """
    配置
    """
    global __cli_args
    parser = ArgumentParser(description=description,
                            epilog=doc, formatter_class=RawTextHelpFormatter)
    parser.add_argument('-v', '--version',
                        action='version', version=version)
    parser.add_argument('-c', '--config', help="run with config file [配置文件路径]")

    # 参数定义
    parser.add_argument('--dns', help="DNS Provider [DNS服务提供商]", choices=[
                        'alidns', 'cloudflare', 'dnscom', 'dnspod', 'dnspod_com', 'he', 'huaweidns', 'callback'])
    parser.add_argument('--id', help="api ID [授权账户]")
    parser.add_argument('--token', help="api token or Secret key [授权访问凭证或密钥]")
    parser.add_argument('--ipv4', nargs="*",
                        help="ipv4 domain list [IPV4域名列表]")
    parser.add_argument('--ipv6', nargs="*",
                        help="ipv6 domain list [IPV6域名列表]")
    parser.add_argument('--index4', help="the way to get ipv4 [IPV4 获取方式]")
    parser.add_argument('--index6', help="the way to get ipv6 [IPV6获取方式]")
    parser.add_argument('--ttl', type=int, help="ttl for DNS [DNS 解析 TTL 时间]")
    parser.add_argument('--proxy', nargs="*",
                        help="https proxy [设置http 代理，多代理逐个尝试直到成功]")
    parser.add_argument('--debug',  type=str2bool, nargs='?',
                        const=True, help="debug mode [是否开启调试,默认否]", )
    parser.add_argument('--cache',  type=str2bool, nargs='?',
                        const=True, help="enable cache [是否缓存记录,默认是]")

    __cli_args = parser.parse_args()
    is_configfile_optional = get_config("token") or get_config("id")
    config_file = get_config("config")
    if not is_configfile_optional or config_file is not None:
        __load_config(config_file or "config.json", is_configfile_optional)
        __cli_args.config = config_file or "config.json"


def __load_config(path="config.json", skip_auto_generation=False):
    """
    加载配置
    """
    global __config
    try:
        with open(path) as configfile:
            __config = loadjson(configfile)
            __config["config_modified_time"] = stat(path).st_mtime
    except IOError:
        if skip_auto_generation:
            __config["config_modified_time"] = time()
            return
        error(' Config file `%s` does not exist!' % path)
        with open(path, 'w') as configfile:
            configure = {
                "$schema": "https://ddns.newfuture.cc/schema/v2.8.json",
                "id": "YOUR ID or EMAIL for DNS Provider",
                "token": "YOUR TOKEN or KEY for DNS Provider",
                "dns": "dnspod",
                "ipv4": [
                    "newfuture.cc",
                    "ddns.newfuture.cc"
                ],
                "ipv6": [
                    "newfuture.cc",
                    "ipv6.ddns.newfuture.cc"
                ],
                "index4": "default",
                "index6": "default",
                "ttl": None,
                "proxy": None,
                "debug": False,
            }
            dumpjson(configure, configfile, indent=2, sort_keys=True)
            sys.stdout.write(
                "New template configure file `%s` is generated.\n" % path)
            sys.exit(1)
    except Exception:
        sys.exit('fail to load config from file: %s' % path)


def get_config(key, default=None):
    """
    读取配置
    1. 命令行参数
    2. 配置文件
    3. 环境变量
    """
    if hasattr(__cli_args, key) and getattr(__cli_args, key) is not None:
        return getattr(__cli_args, key)
    if key in __config:
        return __config.get(key)
    env_name = 'DDNS_'+key.upper()  # type:str
    if env_name in environ:  # 大写环境变量
        return environ.get(env_name)
    if env_name.lower() in environ:  # 小写环境变量
        return environ.get(env_name.lower())
    return default



# coding=utf-8
"""
Hurricane Electric (he.net) API
Hurricane Electric (he.net) 接口解析操作库
https://dns.he.net/docs.html
@author: NN708
"""

from logging import debug, info, warning

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode

__author__ = 'NN708'


class Config:
    TOKEN = "password"
    PROXY = None  # 代理设置


class API:
    # API 配置
    SITE = "dyn.dns.he.net"
    METHOD = "POST"
    ACTION = "nic/update"
    TOKEN_PARAM = "password"  # key name of token param


def request(param=None, **params):
    """
    发送请求数据
    """
    if param:
        params.update(param)

    params.update({API.TOKEN_PARAM: '***'})
    info("%s/%s : %s", API.SITE, API.ACTION, params)
    params[API.TOKEN_PARAM] = Config.TOKEN

    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)

    conn.request(API.METHOD, '/' + API.ACTION, urlencode(params), {
        "Content-type": "application/x-www-form-urlencoded"
    })
    response = conn.getresponse()
    res = response.read().decode('utf8')
    conn.close()

    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]:%s', API.ACTION, response.status, res)
        raise Exception(res)
    else:
        debug('%s : result:%s', API.ACTION, res)
        if not res:
            raise Exception("empty response")
        elif res[:5] == "nochg" or res[:4] == "good":  # No change or success
            return res
        else:
            raise Exception(res)


def update_record(domain, value, record_type="A"):
    """
    更新记录
    """
    info(">>>>>%s(%s)", domain, record_type)
    res = request(hostname=domain, myip=value)
    if res[:4] == "good":
        result = "Record updated. New IP is: " + res[5:-1]
    elif res[:5] == "nochg":
        result = "IP not changed. IP is: " + res[6:-1]
    else:
        result = "Record update failed."
    return result



# coding=utf-8
"""
CloudFlare API
CloudFlare 接口解析操作库
https://api.cloudflare.com/#dns-records-for-a-zone-properties
@author: TongYifan
"""

from json import loads as jsondecode, dumps as jsonencode
from logging import debug, info, warning

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode

__author__ = 'TongYifan'


class Config:
    ID = "AUTH EMAIL"  # CloudFlare 验证的是用户Email，等同于其他平台的userID
    TOKEN = "API KEY"
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SITE = "api.cloudflare.com"  # API endpoint


def request(method, action, param=None, **params):
    """
        发送请求数据
    """
    if param:
        params.update(param)

    params = dict((k, params[k]) for k in params if params[k] is not None)
    info("%s/%s : %s", API.SITE, action, params)
    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)

    if method in ['PUT', 'POST', 'PATCH']:
        # 从public_v(4,6)获取的IP是bytes类型，在json.dumps时会报TypeError
        params['content'] = str(params.get('content'))
        params = jsonencode(params)
    else:  # (GET, DELETE) where DELETE doesn't require params in Cloudflare
        if params:
            action += '?' + urlencode(params)
        params = None
    if not Config.ID:
        headers = {"Content-type": "application/json",
                   "Authorization": "Bearer " + Config.TOKEN}
    else:
        headers = {"Content-type": "application/json",
                   "X-Auth-Email": Config.ID, "X-Auth-Key": Config.TOKEN}
    conn.request(method, '/client/v4/zones' + action, params, headers)
    response = conn.getresponse()
    res = response.read().decode('utf8')
    conn.close()
    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]:%s', action, response.status, res)
        raise Exception(res)
    else:
        data = jsondecode(res)
        debug('%s : result:%s', action, data)
        if not data:
            raise Exception("Empty Response")
        elif data.get('success'):
            return data.get('result', [{}])
        else:
            raise Exception(data.get('errors', [{}]))


def get_zone_id(domain):
    """
        切割域名获取主域名ID(Zone_ID)
        https://api.cloudflare.com/#zone-list-zones
    """
    zoneid = None
    domain_slice = domain.split('.')
    index = 2
    # ddns.example.com => example.com; ddns.example.eu.org => example.eu.org
    while (not zoneid) and (index <= len(domain_slice)):
        zones = request('GET', '', name='.'.join(domain_slice[-index:]))
        zone = next((z for z in zones if domain.endswith(z.get('name'))), None)
        zoneid = zone and zone['id']
        index += 1
    return zoneid


def get_records(zoneid, **conditions):
    """
           获取记录ID
           返回满足条件的所有记录[]
           TODO 大于100翻页
    """
    cache_key = zoneid + "_" + \
        conditions.get('name', "") + "_" + conditions.get('type', "")
    if not hasattr(get_records, 'records'):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ('id', 'type', 'name', 'content', 'proxied', 'ttl')

    if zoneid not in get_records.records:
        get_records.records[cache_key] = {}
        data = request('GET', '/' + zoneid + '/dns_records',
                       per_page=100, **conditions)
        if data:
            for record in data:
                get_records.records[cache_key][record['id']] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}

    records = {}
    for (zid, record) in get_records.records[cache_key].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[zid] = record
    return records


def update_record(domain, value, record_type="A"):
    """
    更新记录
    """
    info(">>>>>%s(%s)", domain, record_type)
    zoneid = get_zone_id(domain)
    if not zoneid:
        raise Exception("invalid domain: [ %s ] " % domain)

    records = get_records(zoneid, name=domain, type=record_type)
    cache_key = zoneid + "_" + domain + "_" + record_type
    result = {}
    if records:  # update
        # https://api.cloudflare.com/#dns-records-for-a-zone-update-dns-record
        for (rid, record) in records.items():
            if record['content'] != value:
                res = request('PUT', '/' + zoneid + '/dns_records/' + record['id'],
                              type=record_type, content=value, name=domain, proxied=record['proxied'], ttl=Config.TTL)
                if res:
                    get_records.records[cache_key][rid]['content'] = value
                    result[rid] = res.get("name")
                else:
                    result[rid] = "Update fail!\n" + str(res)
            else:
                result[rid] = domain
    else:  # create
        # https://api.cloudflare.com/#dns-records-for-a-zone-create-dns-record
        res = request('POST', '/' + zoneid + '/dns_records',
                      type=record_type, name=domain, content=value, proxied=False, ttl=Config.TTL)
        if res:
            get_records.records[cache_key][res['id']] = res
            result = res
        else:
            result = domain + " created fail!"
    return result
